# coding=utf-8


import unittest
from poco.drivers.android.uiautomation import AndroidUiautomationPoco


class TestNeteaseSdkLogin(unittest.TestCase):
    def test_normal_login(self):
        poco = AndroidUiautomationPoco()
        poco(textMatches='^.*登录方式$').wait_for_appearance()
        poco(textMatches='^.*邮箱.*$').click([0.5, -3])
        poco(nameMatches='^.*:id/netease_mpay__login_urs$').click()
        poco(nameMatches='^.*:id/netease_mpay__login_urs$').set_text('adolli@163.com')
        poco(nameMatches='.*:id/netease_mpay__login_login').click()
        poco(nameMatches='.*:id/netease_mpay__login_password').click()
        poco(nameMatches='.*:id/netease_mpay__login_password').set_text('********')
        poco(nameMatches='.*:id/netease_mpay__login_login').click()


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
__author__ = 'lxn3032'


class ControllerBase(object):
    def __init__(self, period, ValueType=float):
        self.ValueType = ValueType
        self.T = period
        self.error_1 = ValueType(0)
        self.error_2 = ValueType(0)
        self.current_value = ValueType(0)
        self.target_value = ValueType(0)

    def delta_closed_loop_gain(self, feedback):
        raise NotImplementedError

    def close_loop_gain(self, feedback):
        raise NotImplementedError

    def set_target_value(self, val):
        self.target_value = val

    def get_current_value(self):
        return self.current_value

    def reset_errors(self):
        self.error_1 = self.error_2 = self.ValueType(0)


class PIDController(ControllerBase):
    def __init__(self, period, Kp=1, Ki=0, Kd=0, ValueType=float):
        super(PIDController, self).__init__(period, ValueType)
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd
        self.sum_error = ValueType(0)

    def delta_closed_loop_gain(self, feedback):
        self.current_value = feedback
        error = self.target_value - self.current_value
        d_error = error - self.error_1
        d2_error = error - 2 * self.error_1 + self.error_2
        delta_output = self.Kp * d_error + self.Ki * error + self.Kd * d2_error

        self.error_2 = self.error_1
        self.error_1 = error
        return delta_output

    def closed_loop_gain(self, feedback):
        self.current_value = feedback
        error = self.target_value - self.current_value
        self.sum_error += error
        d_error = error - self.error_1
        output = self.Kp * error + self.Ki * self.sum_error + self.Kd * d_error

        self.error_2 = self.error_1
        self.error_1 = error
        return output



# coding=utf-8

from poco.sdk.interfaces.hierarchy import HierarchyInterface
from poco.sdk.interfaces.input import InputInterface
from poco.sdk.interfaces.screen import ScreenInterface
from poco.sdk.interfaces.command import CommandInterface

__author__ = 'lxn3032'


def _assign(val, default_val):
    if isinstance(val, type(None)):
        return default_val
    else:
        return val


class PocoAgent(object):
    """
    This is the agent class for poco to communicate with target device.

    This class is an aggregation of 4 major interfaces for now.

    - :py:class:`HierarchyInterface <poco.sdk.interfaces.hierarchy.HierarchyInterface>`: defines the hierarchy
      accessibility methods such as dump(crawl the whole UI tree), getAttr(retrieve attribute value by name)
    - :py:class:`InputInterface <poco.sdk.interfaces.input.InputInterface>`: defines the simulated input methods to
      allow inject simulated input on target device
    - :py:class:`ScreenInterface <poco.sdk.interfaces.screen.ScreenInterface>`: defines methods to access the screen
      surface
    - :py:class:`CommandInterface <poco.sdk.interfaces.command.CommandInterface>`: defines methods to communicate
      with target device in arbitrary way. This is optional.
    """

    def __init__(self, hierarchy, input, screen, command=None):
        self.hierarchy = _assign(hierarchy, HierarchyInterface())
        self.input = _assign(input, InputInterface())
        self.screen = _assign(screen, ScreenInterface())
        self.command = _assign(command, CommandInterface())

    def get_sdk_version(self):
        pass



# coding=utf-8


import time
import json
from airtest_hunter import AirtestHunter, open_platform
from poco.drivers.netease.internal import NeteasePoco

from pocounit.case import PocoTestCase
from airtest.core.api import connect_device, device as current_device
from poco.drivers.android.uiautomation import AndroidUiautomationPoco


class Case(PocoTestCase):
    @classmethod
    def setUpClass(cls):
        super(Case, cls).setUpClass()
        if not current_device():
            connect_device('Android:///')

    def runTest(self):
        from poco.drivers.cocosjs import CocosJsPoco
        poco = CocosJsPoco()
        for n in poco():
            print n.get_name()


# if __name__ == '__main__':
#     import pocounit
#     pocounit.main()


# from hunter_cli import Hunter, open_platform
# from poco.drivers.netease.internal import NeteasePoco
#
# tokenid = open_platform.get_api_token('test')
# hunter = Hunter(tokenid, 'xy2', 'xy2_at_408d5c116536')
# poco = NeteasePoco('xy2', hunter)
#
# print poco('npc_conversation').offspring('list_options').offspring('Widget')[0].offspring('txt_content').nodes[0].node.data

from airtest.core.api import connect_device
from poco.utils.track import track_sampling, MotionTrack, MotionTrackBatch
from poco.utils.airtest.input import AirtestInput


mt = MotionTrack()
mt1 = MotionTrack()
mt2 = MotionTrack()
mt.start([0.5, 0.5]).move([0.2, 0.5]).move([0.5, 0.5]).hold(1)
mt1.start([0.5, 0.6]).move([0.2, 0.6]).hold(1).move([0.5, 0.6])
# mt2.hold(1).start([0.5, 0.4]).move([0.2, 0.4]).move([0.5, 0.4])


meb = MotionTrackBatch([mt1, mt])
for e in meb.discretize():
    print e
print len(meb.discretize())

connect_device('Android:///')
poco = AndroidUiautomationPoco(use_airtest_input=True)

# ai = AirtestInput()
# t1 = time.time()
# ai.applyMotionTracks(meb.discretize())
# t2 = time.time()
# print t2 - t1

poco.apply_motion_tracks([mt1, mt])

time.sleep(4)




# encoding=utf-8
import sys
sys.path.append("../../..")
from simplerpc.rpcserver import RpcServer
from simplerpc.simplerpc import dispatcher, AsyncResponse
import time


@dispatcher.add_method
def foobar(**kwargs):
    return kwargs["foo"] + kwargs["bar"]


@dispatcher.add_method
def make_error(*args):
    raise


@dispatcher.add_method
def delayecho(*args):
    r = AsyncResponse()
    from threading import Thread

    def func(r):
        time.sleep(5)
        r.result(args)

    Thread(target=func, args=(r,)).start()
    return r


@dispatcher.add_method
def delayerror(*args):
    r = AsyncResponse()
    from threading import Thread

    def func(r):
        time.sleep(5)
        r.error(RuntimeError("something wrong here"))

    Thread(target=func, args=(r,)).start()
    return r


def test_with_tcp():
    from simplerpc.transport.tcp import TcpServer
    s = RpcServer(TcpServer())
    s.run()
    # s.console_run({"s": s})


if __name__ == '__main__':
    test_with_tcp()



# coding=utf-8

from airtest_hunter import open_platform, AirtestHunter
from hunter_cli.rpc.client import HunterRpcClient

from poco import Poco
from poco.agent import PocoAgent
from poco.utils.airtest.input import AirtestInput
from poco.utils.airtest.screen import AirtestScreen
from poco.utils.hrpc.hierarchy import RemotePocoHierarchy
from poco.utils.hunter.command import HunterCommand

__all__ = ['NeteasePoco']
__author__ = 'lxn3032'


class NeteasePocoAgent(PocoAgent):
    def __init__(self, hunter):
        client = HunterRpcClient(hunter)
        client.set_timeout(25)
        remote_poco = client.remote('poco-uiautomation-framework-2')

        # hierarchy
        dumper = remote_poco.dumper
        selector = remote_poco.selector
        attributor = remote_poco.attributor
        hierarchy = RemotePocoHierarchy(dumper, selector, attributor)

        # input
        input = AirtestInput()

        # screen
        screen = AirtestScreen()

        # command
        command = HunterCommand(hunter)

        super(NeteasePocoAgent, self).__init__(hierarchy, input, screen, command)
        self._rpc_client = client


class NeteasePoco(Poco):
    def __init__(self, process, hunter=None, **options):
        apitoken = open_platform.get_api_token(process)
        self._hunter = hunter or AirtestHunter(apitoken, process)
        agent = NeteasePocoAgent(self._hunter)
        super(NeteasePoco, self).__init__(agent, **options)
        self._last_proxy = None

    def on_pre_action(self, action, proxy, args):
        try:
            from airtest.core.api import snapshot
        except ImportError:
            # 兼容旧airtest
            from airtest.core.main import snapshot
            snapshot(msg=unicode(proxy))



# coding=utf-8
from __future__ import unicode_literals

import time
import traceback
import warnings

from .acceleration import PocoAccelerationMixin
from .exceptions import PocoTargetTimeout, InvalidOperationException
from .proxy import UIObjectProxy
from .agent import PocoAgent
from .freezeui.utils import create_immutable_hierarchy
from .utils.track import MotionTrackBatch

__author__ = 'lxn3032'


class Poco(PocoAccelerationMixin):
    """
    Poco standard initializer.

    Args:
        agent (:py:class:`PocoAgent <poco.agent.PocoAgent>`): an agent object for Poco to communicate with the target
         device. See :py:class:`PocoAgent <poco.agent.PocoAgent>` definition for more details.
        options:
            - ``action_interval``: time interval to wait for the action (such as touch or swipe) completion performed
              on device and for the UI to become still (stable). Default value is 0.8s.
            - ``poll_interval``: the minimum time needed between each poll events (such as waiting for UI element to
              appear on the screen). Polling is done periodically.
            - ``pre_action_wait_for_appearance``: time interval to wait before the action (such as click or swipe) is
              performed. If the target UI element does not appear on the screen after this time interval, the
              :py:class:`PocoNoSuchNodeException <poco.exceptions.PocoNoSuchNodeException>` is raised
            - ``touch_down_duration``: Touch down step duration of the click operation last for. If this argument is
              provided, this value will set to ``self.agent.input`` module. Note that not all implementation of poco 
              support this parameter. If not support, you may see a warning.
    """

    def __init__(self, agent, **options):
        super(Poco, self).__init__()
        self._agent = agent

        # options
        self._pre_action_wait_for_appearance = options.get(b'pre_action_wait_for_appearance', 6)
        self._post_action_interval = options.get(b'action_interval', 0.8)
        self._poll_interval = options.get(b'poll_interval', 1.44)
        if 'touch_down_duration' in options:
            touch_down_duration = options[b'touch_down_duration']
            try:
                touch_down_duration = float(touch_down_duration)
            except ValueError:
                raise ValueError('Option `touch_down_duration` should be <float>. Got {}'
                                 .format(repr(touch_down_duration)))
            self._agent.input.setTouchDownDuration(touch_down_duration)

        self._pre_action_callbacks = [self.on_pre_action.__func__]
        self._post_action_callbacks = [self.on_post_action.__func__]

    def __call__(self, name=None, **kw):
        """
        Call Poco instance to select the UI element by query expression. Query expression can contain specific name
        and/or other attributes. Invisible UI elements will be skipped even if "visible=False" argument is set.

        Selection process is not executed instantly, the query expression is stored in the UI proxy and the selection is
        executed only then when the UI element(s) info is required (such get the point coordinates where to click, 
        and/or retrieve the specific attribute value).

        Examples:
            This example shows selecting a Button named 'close'::

                poco = Poco(...)
                close_btn = poco('close', type='Button')

        Args:
            name (:obj:`str`): name of the UI element to be selected

        Keyword Args:
            xx: arbitrary key value pair that stands for selecting the UI matching the value of ``UI.xx``
            xxMatches (:obj:`str`): arbitrary key value pair that stands for selecting the UI matching the regular 
             expression pattern ``UI.xx``

        In keyword args, you can only use `xx` or `xxMatches` at the same time. Using both with the same attribute does
        not make sense. Besides, `xx` should not start with ``_`` (underscore) as attributes start with ``_`` are 
        private attributes that used by sdk implementation.
        ::

            # select the UI element(s) which text attribute matches the pattern '^close.*$'
            poco = Poco(...)
            arb_close_btn = poco(textMatches='^close.*$')

        Returns:
            :py:class:`UIObjectProxy <poco.proxy.UIObjectProxy>`: UI proxy object representing the UI element matches 
            the given query expression.
        """

        if not name and len(kw) == 0:
            warnings.warn("Wildcard selector may cause performance trouble. Please give at least one condition to "
                          "shrink range of results")
        return UIObjectProxy(self, name, **kw)

    def wait_for_any(self, objects, timeout=120):
        """
        Wait until any of given UI proxies show up before timeout and return the first appeared UI proxy.
        All UI proxies will be polled periodically. See options :py:class:`poll_interval <poco.pocofw.Poco>` in
        ``Poco``'s initialization for more details.

        Args:
            objects (Iterable<:py:class:`UIObjectProxy <poco.proxy.UIObjectProxy>`>): iterable object of the given UI 
             proxies
            timeout (:obj:`float`): timeout in seconds, default is 120s

        Returns:
            :py:class:`UIObjectProxy <poco.proxy.UIObjectProxy>`: the first appeared UI proxy

        Raises:
            PocoTargetTimeout: when none of UI proxies appeared before timeout
        """

        start = time.time()
        while True:
            for obj in objects:
                if obj.exists():
                    return obj
            if time.time() - start > timeout:
                raise PocoTargetTimeout('any to appear', repr(objects).decode('utf-8'))
            self.sleep_for_polling_interval()

    def wait_for_all(self, objects, timeout=120):
        """
        Wait until all of given UI proxies show up before timeout.
        All UI proxies will be polled periodically. See option :py:class:`poll_interval <poco.pocofw.Poco>` in 
        ``Poco``'s initialization for more details.

        Args:
            objects (Iterable<:py:class:`UIObjectProxy <poco.proxy.UIObjectProxy>`>): iterable object of the given UI 
             proxies
            timeout (:obj:`float`): timeout in seconds, default is 120s

        Raises:
            PocoTargetTimeout: when not all of UI proxies appeared before timeout
        """

        start = time.time()
        while True:
            all_exist = True
            for obj in objects:
                if not obj.exists():
                    all_exist = False
                    break
            if all_exist:
                return
            if time.time() - start > timeout:
                raise PocoTargetTimeout('all to appear', repr(objects).decode('utf-8'))
            self.sleep_for_polling_interval()

    def freeze(this):
        """
        Snapshot current **hierarchy** and cache it into a new poco instance. This new poco instance is a copy from
        current poco instance (``self``). The hierarchy of the new poco instance is fixed and immutable. It will be
        super fast when calling ``dump`` function from frozen poco. See the example below.
        
        Examples:
            ::

                poco = Poco(...)
                frozen_poco = poco.freeze()
                hierarchy_dict = frozen_poco.agent.hierarchy.dump()  # will return the already cached hierarchy data
                

        Returns:
            :py:class:`Poco <poco.pocofw.Poco>`: new poco instance copy from current poco instance (``self``)
        """

        class FrozenPoco(Poco):
            def __init__(self):
                hierarchy_dict = this.agent.hierarchy.dump()
                hierarchy = create_immutable_hierarchy(hierarchy_dict)
                agent_ = PocoAgent(hierarchy, this.agent.input, this.agent.screen)
                super(FrozenPoco, self).__init__(agent_, action_interval=0.01, pre_action_wait_for_appearance=0)
                self._pre_action_callbacks = this._pre_action_callbacks
                self._post_action_callbacks = this._post_action_callbacks

            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                pass

        return FrozenPoco()

    def wait_stable(self):
        """
        Sleep for fixed number of seconds in order to wait for the UI to become still (stable).
        There is no need to call this method manually. It's automatically invoked when required.
        """

        time.sleep(self._post_action_interval)

    def sleep_for_polling_interval(self):
        """
        Sleep for fixed number of seconds after each poll event.
        There is no need to call this method manually. It's automatically invoked when required.
        """

        time.sleep(self._poll_interval)

    @property
    def agent(self):
        """
        Readonly property to access poco agent instance. See :py:class:`poco.agent.PocoAgent` for more details.

        Returns:
            :py:class:`poco.agent.PocoAgent`: poco agent instance
        """

        return self._agent

    def click(self, pos):
        """
        Perform click (touch, tap, etc.) action on target device at given coordinates.

        The coordinates (x, y) are either a 2-list or 2-tuple. The coordinates values for x and y must be in the
        interval between 0 ~ 1 to represent the percentage of the screen. For example, the coordinates ``[0.5, 0.5]``
        represent the `center` of the screen and the coordinates ``[0, 0]`` represent the `top left corner`.

        See ``CoordinateSystem`` for more details about coordinate system.

        Examples:
            Click the point of ``(100, 100)`` of screen which resolution is ``(1920, 1080)``::

                poco.click([100.0 / 1920, 100.0 / 1080])

        Args:
            pos (:obj:`list(float, float)` / :obj:`tuple(float, float)`): coordinates (x, y) in range of 0 to 1

        Raises:
            InvalidOperationException: when clicked outside of the screen
        """

        if not (0 <= pos[0] <= 1) or not (0 <= pos[1] <= 1):
            raise InvalidOperationException('Click position out of screen. pos={}'.format(repr(pos).decode('utf-8')))
        return self.agent.input.click(pos[0], pos[1])

    def swipe(self, p1, p2=None, direction=None, duration=2.0):
        """
        Perform swipe action on target device from point to point given by start point and end point, or by the
        direction vector. At least one of the end point or direction must be provided.

        The coordinates (x, y) definition for points is the same as for ``click`` event. The components of the
        direction vector (x, y) are also expressed in the range of the screen from 0 to 1.

        See ``CoordinateSystem`` for more details about coordinate system.

        Examples:
            Following example shows how to perform a swipe action from (100, 100) to (100, 200) on screen with
            resolution 1920x1080::

                poco.swipe([100.0 / 1920, 100.0 / 1080], [100.0 / 1920, 200.0 / 1080])

            Or given by the specific direction instead of end point::

                poco.swipe([100.0 / 1920, 100.0 / 1080], direction=[0, 100.0 / 1080])

        Args:
            p1 (:obj:`2-list/2-tuple`): start point
            p2: end point
            direction: swipe direction
            duration (:obj:`float`): time interval in which the swipe action is performed

        Raises:
            InvalidOperationException: when the start point of the swipe action lies outside the screen
        """

        try:
            duration = float(duration)
        except ValueError:
            raise ValueError('Argument `duration` should be <float>. Got {}'.format(repr(duration)))

        if not (0 <= p1[0] <= 1) or not (0 <= p1[1] <= 1):
            raise InvalidOperationException('Swipe origin out of screen. {}'.format(repr(p1).decode('utf-8')))
        if direction is not None:
            p2 = [p1[0] + direction[0], p1[1] + direction[1]]
        elif p2 is not None:
            p2 = p2
        else:
            raise TypeError('Swipe end not set.')
        return self.agent.input.swipe(p1[0], p1[1], p2[0], p2[1], duration)

    def long_click(self, pos, duration=2.0):
        """
        Similar to click but press the screen for the given time interval and then release

        Args:
            pos (:obj:`2-list/2-tuple`): coordinates (x, y) in range from 0 to 1
            duration: duration of press the screen
        """

        try:
            duration = float(duration)
        except ValueError:
            raise ValueError('Argument `duration` should be <float>. Got {}'.format(repr(duration)))

        if not (0 <= pos[0] <= 1) or not (0 <= pos[1] <= 1):
            raise InvalidOperationException('Click position out of screen. {}'.format(repr(pos).decode('utf-8')))
        return self.agent.input.longClick(pos[0], pos[1], duration)

    def apply_motion_tracks(self, tracks, accuracy=0.004):
        """
        Similar to click but press the screen for the given time interval and then release

        Args:
           tracks (:py:obj:`list`): list of :py:class:`poco.utils.track.MotionTrack` object
           accuracy (:py:obj:`float`): motion accuracy for each motion steps in normalized coordinate metrics.
        """

        if not tracks:
            raise ValueError('Please provide at least one track. Got {}'.format(repr(tracks)))

        tb = MotionTrackBatch(tracks)
        return self.agent.input.applyMotionTracks(tb.discretize(accuracy))

    def snapshot(self, width=720):
        """
        Take the screenshot from the target device. The supported output format (png, jpg, etc.) depends on the agent
        implementation.

        Args:
            width (:obj:`int`): an expected width of the screenshot. The real size depends on the agent implementation
            and might not be possible to obtain the expected width of the screenshot

        Returns:
            2-tuple:
                - screen_shot (:obj:`str/bytes`): base64 encoded screenshot data
                - format (:obj:`str`): output format 'png', 'jpg', etc.
        """

        return self.agent.screen.getScreen(width)

    def get_screen_size(self):
        """
        Get the real physical resolution of the screen of target device.

        Returns:
            tuple: float number indicating the screen physical resolution in pixels
        """

        return self.agent.screen.getPortSize()

    def command(self, cmd, type=None):
        return self.agent.command.command(cmd, type)

    def on_pre_action(self, action, proxy, args):
        pass

    def on_post_action(self, action, proxy, args):
        pass

    def add_pre_action_callback(self, cb):
        """
        Register a callback function to be invoked before each action (such as touch or swipe).

        The callback function arguments are defined as follows:

        * ``action`` (:obj:`str`): name or tag of the action
        * ``proxy`` (:py:class:`UIObjectProxy <poco.proxy.UIObjectProxy>` or :obj:`NoneType`): related UI proxy which is
          involved in the action itself
        * ``args`` (:obj:`tuple`): all required arguments of the specific action function

        Args:
            cb: the callback function
        """

        self._pre_action_callbacks.append(cb)

    def add_post_action_callback(self, cb):
        """
        Register a callback function to be invoked after each action (such as touch or swipe).

        The arguments to be passed are identical to the callback function in
        :py:meth:`add_pre_action_callback <poco.pocofw.Poco.add_pre_action_callback>`.

        Args:
            cb: the callback function
        """

        self._post_action_callbacks.append(cb)

    def pre_action(self, action, proxy, args):
        for cb in self._pre_action_callbacks:
            try:
                cb(self, action, proxy, args)
            except:
                warnings.warn("Error occurred at pre action stage.\n{}".format(traceback.format_exc()))

    def post_action(self, action, proxy, args):
        for cb in self._post_action_callbacks:
            try:
                cb(self, action, proxy, args)
            except:
                warnings.warn("Error occurred at post action stage.\n{}".format(traceback.format_exc()))



from ..dispatcher import Dispatcher
import sys
if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class Math:

    def sum(self, a, b):
        return a + b

    def diff(self, a, b):
        return a - b


class TestDispatcher(unittest.TestCase):

    """ Test Dispatcher functionality."""

    def test_getter(self):
        d = Dispatcher()

        with self.assertRaises(KeyError):
            d["method"]

        d["add"] = lambda *args: sum(args)
        self.assertEqual(d["add"](1, 1), 2)

    def test_in(self):
        d = Dispatcher()
        d["method"] = lambda: ""
        self.assertIn("method", d)

    def test_add_method(self):
        d = Dispatcher()

        @d.add_method
        def add(x, y):
            return x + y

        self.assertIn("add", d)
        self.assertEqual(d["add"](1, 1), 2)

    def test_add_class(self):
        d = Dispatcher()
        d.add_class(Math)

        self.assertIn("math.sum", d)
        self.assertIn("math.diff", d)
        self.assertEqual(d["math.sum"](3, 8), 11)
        self.assertEqual(d["math.diff"](6, 9), -3)

    def test_add_object(self):
        d = Dispatcher()
        d.add_object(Math())

        self.assertIn("math.sum", d)
        self.assertIn("math.diff", d)
        self.assertEqual(d["math.sum"](5, 2), 7)
        self.assertEqual(d["math.diff"](15, 9), 6)

    def test_add_dict(self):
        d = Dispatcher()
        d.add_dict({"sum": lambda *args: sum(args)}, "util")

        self.assertIn("util.sum", d)
        self.assertEqual(d["util.sum"](13, -2), 11)

    def test_add_method_keep_function_definitions(self):

        d = Dispatcher()

        @d.add_method
        def one(x):
            return x

        self.assertIsNotNone(one)

    def test_del_method(self):
        d = Dispatcher()
        d["method"] = lambda: ""
        self.assertIn("method", d)

        del d["method"]
        self.assertNotIn("method", d)

    def test_to_dict(self):
        d = Dispatcher()
        func = lambda: ""
        d["method"] = func
        self.assertEqual(dict(d), {"method": func})

    def test_init_from_object_instance(self):

        class Dummy():

            def one(self):
                pass

            def two(self):
                pass

        dummy = Dummy()

        d = Dispatcher(dummy)

        self.assertIn("one", d)
        self.assertIn("two", d)
        self.assertNotIn("__class__", d)

    def test_init_from_dictionary(self):

        dummy = {
            'one': lambda x: x,
            'two': lambda x: x,
        }

        d = Dispatcher(dummy)

        self.assertIn("one", d)
        self.assertIn("two", d)

    def test_dispatcher_representation(self):

        d = Dispatcher()
        self.assertEqual('{}', repr(d))



# coding=utf-8
from __future__ import unicode_literals

import poco.utils.six as six


__all__ = ['query_expr']


TranslatePred = {
    'attr=': '=',
    'attr.*=': ' matches ',
}


TranslateOp = {
    'and': '&',
    'or': '|',
    '/': '/',
    '>': '>',
    '-': '-',
}


def query_expr(query):
    op = query[0]
    if op in ('/', '>', '-'):
        return TranslateOp[op].join([query_expr(q) for q in query[1]])
    elif op == 'index':
        return '{}[{}]'.format(query_expr(query[1][0]), query[1][1])
    elif op in ('and', 'or'):
        exprs = []
        for subquery in query[1]:
            pred, (k, v) = subquery
            if k == 'name':
                exprs.append(v)
            else:
                exprs.append('{}{}{}'.format(k, TranslatePred[pred], v))
        return TranslateOp[op].join(exprs)
    else:
        raise RuntimeError('Bad query format. "{}"'.format(repr(query)))


def ensure_string_value(value):
    if six.PY2 and isinstance(value, str):
        return value.decode("utf-8")
    else:
        return value


def build_query(name, **attrs):
    query = []
    if name is not None:
        name = ensure_string_value(name)
        attrs['name'] = name
    for attr_name, attr_val in attrs.items():
        attr_val = ensure_string_value(attr_val)
        if attr_name.startswith('_'):
            raise NameError("Cannot use private attribute '{}' in your Query Expression as private attributes do not "
                            "have stable values.".format(attr_name))
        elif attr_name.endswith('Matches'):
            attr_name = attr_name[:-7]  # textMatches -> (attr.*=, text)
            op = 'attr.*='
        else:
            op = 'attr='
        query.append((op, (attr_name, attr_val)))
    return 'and', tuple(query)
